#include "slideface.h"

#include "slideworld.h"
#include "slidepoint.h"
#include "slidesurface.h"

BOOL CSLIDEFace::m_bBackFaceCullingEnabled = TRUE;
BOOL CSLIDEFace::m_bWrongBackFaceCullingTestEnabled = FALSE;
BOOL CSLIDEFace::m_bClipEnabled = TRUE;
BOOL CSLIDEFace::m_bRenderNormalsEnabled = FALSE;
FLOAT CSLIDEFace::m_fNormalsRenderLength = (FLOAT)0.2;

CListInt *CSLIDEFace::m_plistVertices = NULL;
UINT_32 CSLIDEFace::m_uiCount = 0;

//////////////////////////////////////////////////////////////////////////////
// CSLIDEFace Class
//

CSLIDEFace::CSLIDEFace()
{
  m_uiCount++;
  if ( m_plistVertices == NULL )
    {
      m_plistVertices = new CListInt;
      ASSERT( m_plistVertices != NULL );
      m_plistVertices->Init();
    }

  m_pcSurfaceID = NULL;
  m_pCSLIDESurface = NULL;
  m_plistPointIDs = NULL;
}

CSLIDEFace::~CSLIDEFace()
{
  m_uiCount--;
  if ( m_uiCount == 0 )
    {
      if ( m_plistVertices != NULL )
	{
	  delete m_plistVertices;
	  m_plistVertices = NULL;
	}
    }
}

VOID CSLIDEFace::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID, 
		      CListInt *plistPointIDs)
{
  CSLIDEPrimitive::Init(pcFile, uiLine, uiColumn, pcID);

  ASSERT( plistPointIDs != NULL );

  m_pcSurfaceID = gcppcSLIDEReservedWords[sliderwtSLF_INHERIT];
  m_pCSLIDESurface = NULL;

  m_plistPointIDs = plistPointIDs;
  m_listCSLIDEPoint.Init();
}

VOID CSLIDEFace::Uninit()
{
  CSLIDEPrimitive::Uninit();

  if ( m_plistPointIDs != NULL )
    {
      delete m_plistPointIDs;
      m_plistPointIDs = NULL;
    }
}

BOOL CSLIDEFace::Initialized()
{
  if ( !CSLIDEPrimitive::Initialized() )
    {
      return FALSE;
    }

  return ( m_pcSurfaceID != NULL &&
	   m_plistPointIDs != NULL );
}

BOOL CSLIDEFace::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect;
  CListIterInt iter;
  CHAR *pcID;
  CSLIDEPoint *pCSLIDEPoint;

  if ( !CSLIDEPrimitive::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  // Check Surface Reference

  m_pCSLIDESurface = pCSLIDEWorld->PeekSurface(m_pcSurfaceID);
  if ( m_pCSLIDESurface == NULL )
    {
      HandleError("Undefined %s \"%s\"", 
		  gcppcSLIDEReservedWords[sliderwtSurface],
		  m_pcSurfaceID);
      bCorrect = FALSE;
    }

  // Check Point References

  iter.Init(m_plistPointIDs);

  for ( pcID = (CHAR*)(iter.PeekFirst());
	pcID != NULL;
	pcID = (CHAR*)(iter.PeekNext()) )
    {
      pCSLIDEPoint = pCSLIDEWorld->PeekPoint(pcID);
      if ( pCSLIDEPoint == NULL )
	{
	  HandleError("Undefined %s \"%s\"", gcppcSLIDEReservedWords[sliderwtPoint], 
		      pcID);
	  bCorrect = FALSE;
	}
      m_listCSLIDEPoint.InsertLast(pCSLIDEPoint, 0);
    }

  return bCorrect;
}

BOOL CSLIDEFace::Valid()
{
  if ( !CSLIDEPrimitive::Valid() )
    {
      return FALSE;
    }

  return (GetPointCount() >= 3);
}

CSLIDEPrimitiveIter *CSLIDEFace::AllocateIter()
{
  CSLIDEFaceIter *pCSLIDEFaceIter;

  pCSLIDEFaceIter = new CSLIDEFaceIter;
  ASSERT(pCSLIDEFaceIter != NULL );
  pCSLIDEFaceIter->Init(this);

  return pCSLIDEFaceIter;
}

VOID CSLIDEFace::FreeIter(CSLIDEPrimitiveIter *pCSLIDEFaceIter)
{
  delete pCSLIDEFaceIter;
}

VOID CSLIDEFace::InsertIntoBound(CBound *pCBound)
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;

  ASSERT( pCBound != NULL );

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      pCSLIDEPoint->InsertIntoBound(pCBound);
    }
}

VOID CSLIDEFace::Preprocess()
{
  CheckDynamic();
  UpdateState();

  // All faces must calculate their normal at least once

  CalculateNormal();
  CalculateCentroid();
  
  // Calculate the D coefficient of the Tangent Plane defined by the normal
  m_vNormal.CompletePlaneEquation(m_ptCentroid);
}

VOID CSLIDEFace::CheckDynamic()
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;
  BOOL bBounded;
  BOOL bBoundDynamic;

  bBounded = GetDynamic();
  bBoundDynamic = GetDynamic();

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      bBounded = TRUE;
      if ( pCSLIDEPoint->GetBoundDynamic() )
	{
	  bBoundDynamic = TRUE;
	  break;
	}
    }

  SetBounded(bBounded);
  SetBoundDynamic(bBoundDynamic);
  SetDynamic(bBoundDynamic);
  ASSERT( !(!GetBounded() && GetBoundDynamic()) );
}

VOID CSLIDEFace::UpdateState()
{
  if ( GetBoundDynamic() )
    {
      CalculateNormal();
      CalculateCentroid();

      // Calculate the D coefficient of the Tangent Plane defined by the normal
      m_vNormal.CompletePlaneEquation(m_ptCentroid);
    }
}

// This uses Newell's Algorithm for finding the best fit planar equation
// for a set of points in 3 space (pg. 476 Foley and van Dam)
// I modified it so that the normal would be positive for counter-clockwise
// polygons, and I removed the 1/2 coefficient because we are concerned 
// with relative areas, not absolute areas.

VOID CSLIDEFace::CalculateNormal()
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;
  Point ptStart, ptPrev, ptCurr;

  m_vNormal.Null();

  iter.Init(&m_listCSLIDEPoint);

  pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
  if ( pCSLIDEPoint == NULL )
    {
      HandleWarning("Null normal because no points for face");
      return;
    }
  ptStart = pCSLIDEPoint->GetPoint();
  ptPrev = ptStart;

  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      ptCurr = pCSLIDEPoint->GetPoint();

      m_vNormal[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
      m_vNormal[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
      m_vNormal[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);

      ptPrev = ptCurr;
    }

  ptCurr = ptStart;
  m_vNormal[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
  m_vNormal[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
  m_vNormal[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);
  
  if ( !m_vNormal.Normalize() )
    {
      HandleWarning("Null normal because of zero area face");
    }
}

VOID CSLIDEFace::CalculateCentroid()
{
  CListIterInt iter;
  UINT_32 i;
  CSLIDEPoint *pCSLIDEPoint;

  iter.Init(&m_listCSLIDEPoint);

  m_ptCentroid.Null();

  i = 0;
  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst();
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      m_ptCentroid = m_ptCentroid + pCSLIDEPoint->GetPoint();
      i++;
    }

  if ( i > 0 )
    {
      m_ptCentroid = m_ptCentroid / (FLOAT)i;
    }

  m_ptCentroid[W] = 1.0;;
}

//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 6: What about Homogenizing?
//                      Read and understand the back face culling as well! 
//
VOID CSLIDEFace::Render(const Matrix &mGvp_proj, 
			const Matrix &mGproj_vrc, 
			const Matrix &mGvrc_proj, 
			const Matrix &mGvrc_obj, 
			const Matrix &mGobj_vrc, 
			CBound *pCBoundClip, 
			CListInt * /*plistCSLIDELightPath*/,
			CSLIDESurface *pCSLIDESurface,
			SLIDEShadingType /*eShading*/, SLIDELODType eLOD, 
			SLIDESolidType eSolid)
{
  GLenum mode;
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;
  Matrix mGproj_obj;
  Vertex *pvx;

  mGproj_obj = mGproj_vrc * mGvrc_obj;

  // Check Back Face Culling
  if ( GetBackFaceCullingEnabled() && eSolid == slidestSLF_SOLID )
    {
      Vector vNormal;

      if ( GetWrongBackFaceCullingTestEnabled() )
	{
	  // This is if we naively apply the modelling transform to the 
	  // normal direction vector.  This works for most parallel projection
	  // examples, but it fails quickly with the perspective projection.
	  vNormal = m_vNormal;
	  vNormal[3] = 0.0;
	  vNormal = mGproj_obj * vNormal;
	}
      else
	{
	  FLOAT fDeterminant;
	  
	  //This is the correct line
	  // Doing row multiplication instead of transposing the matrix
	  vNormal = m_vNormal * mGobj_vrc * mGvrc_proj; 
	  fDeterminant = mGproj_obj.Determinant();
	  ASSERT( fDeterminant != (FLOAT)0.0 );
	  if ( fDeterminant < (FLOAT)0.0 )
	    {
	      vNormal = -vNormal;
	    }
	}
      
      if ( vNormal[Z] <= 0.0 )
	{
	  return;
	}
    }

  // Modify Rendering Properties
  ASSERT( eLOD == slidelodtSLF_EDGES || eLOD == slidelodtSLF_FULL );
  ModifySurfacePointer(pCSLIDESurface);

  // Build a list of Vertices which we dynamically allocate
  // this will be the input to the clipping routine
  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      pvx = CBound::AllocateVertex();
      ASSERT( pvx != NULL );

      pvx->SetPoint() = mGproj_obj * pCSLIDEPoint->GetPoint();

      m_plistVertices->InsertLast(pvx, 0);
    }

  // Clip the polygon against the 4D bound if clipping is enabled
  // and we do not lie entirely within the camera's view frustum
  if ( GetClipEnabled() && pCBoundClip != NULL )
    {
      // The clipping routine should modify the list of vertices.
      // It should deallocate any vertices it removes by calling 
      // CBound::DeallocateVertex(pvx) on them.  It should also
      // insert any new vertices into the m_plistVertices, so that
      // when the ClipPolygon routine finishes the list is the clipped
      // polygon contour.
      pCBoundClip->ClipPolygon(*m_plistVertices);
    }

  // Transform the clipped vertices into pixel coords, homogenize them, and 
  // render them. We remove the vertices from the list of vertices and 
  // deallocate these temporary vertex structures as we go.
  gilColor3fv(pCSLIDESurface->GetColor().Get());
  if ( eLOD == slidelodtSLF_EDGES )
    {
      mode = GL_LINE_LOOP;
    }
  else
    {
      mode = GL_POLYGON;
    }
  gilBegin(mode);
  {
    //printf("***** Begin Face *****\n");
    for ( pvx = (Vertex*)m_plistVertices->RemoveFirst(); 
	  pvx != NULL;
	  pvx = (Vertex*)m_plistVertices->RemoveFirst() )
      {
	pvx->SetPoint() = mGvp_proj * pvx->GetPoint();
	
	pvx->Render();
	CBound::DeallocateVertex(pvx);
      }
    //printf("***** End Face *****\n");
  }
  gilEnd();

  // Render the normal if desired
  if ( GetRenderNormalsEnabled() )
    {
      gilPushAttrib(GL_CURRENT_BIT);
      {
	gilColor3f(1.0, 0.0, 1.0);
	gilBegin(GL_LINES);
	{
	  RenderNormal(mGvp_proj, mGproj_vrc, mGvrc_obj, mGobj_vrc, 
		       m_ptCentroid, m_vNormal);
	}
	gilEnd();
      }
      gilPopAttrib();
    }
}
//
// END  : ASSIGNMENT 6: What about Homogenizing?
//                      Read and understand the back face culling as well! 
//////////////////////////////////////////////////////////////////////////////

VOID CSLIDEFace::ModifySurfacePointer(CSLIDESurface* &pCSLIDESurface)
{
  if ( m_pCSLIDESurface != CSLIDESurface::GetSLF_INHERIT() )
    {
      pCSLIDESurface = m_pCSLIDESurface;
    }
}

//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 6: Read and understand this!!!
// 

// When viewing the normals with the "Crazy Clipped Rotate" enabled they
// will look strange.  This is because we are not correctly transforming 
// the normal vector.  Instead we are correctly transforming the vector
// into VRC space, and constructing a magenta line segment to represent
// normal vector.  We then apply the Gproj<-vrc transform to the endpoints
// of that line segment.  This is not the same as transforming the normal
// vector.  We do it this way so that the normals of distant objects will
// be forshortened like the geometry of those objects.
//
// A good example of how transforming a line segment is different from 
// transforming a normal is an axially aligned cube.  The line segments 
// representing the normals of the front and back faces will transform 
// correctly.  But the line segments representing the normals of the four
// faces which are parallel to the Z-axis will be very wrong.  The normals
// to these faces line in planes which are parallel to the XY-plane.  We 
// know that the perspective warp makes planes parallel to the XY-plane
// remain that way.  Consequently, those line segments representing the 
// normals will not rotate direction at all.  But the actual faces of the
// cube become slanted due to the perspective warp, which implies that their
// normals should be rotated some amount.  This is why the rendered normals
// will look funny when step out of the body of the viewer and watch what 
// is happening in his perspective projection (using the "Crazy Clipped Rotate"
// rendering option).

VOID CSLIDEFace::RenderNormal(const Matrix &mGvp_proj, const Matrix &mGproj_vrc, 
			      const Matrix &mGvrc_obj, const Matrix &mGobj_vrc,
			      const Point &pt, const Vector &vNormal)
{
  Point ptStart, ptEnd;
  Vector v;
  Matrix mGvp_vrc;

  mGvp_vrc = mGvp_proj * mGproj_vrc;

  v = vNormal * mGobj_vrc;
  ptStart = mGvrc_obj * pt;

  v[W] = 0;
  v.Normalize();
  ptEnd = ptStart + GetNormalsRenderLength() * v;

  // We break here because we want to see the shortening effects of the 
  // perspective projection on the normal

  ptStart = mGvp_vrc * ptStart;
  ptEnd = mGvp_vrc * ptEnd;

  ptStart.Homogenize();
  ptEnd.Homogenize();

  gilVertex3fv(ptStart.Get());
  gilVertex3fv(ptEnd.Get());
}
// 
// END  : ASSIGNMENT 6: Read and understand this!!!
//////////////////////////////////////////////////////////////////////////////

VOID CSLIDEFace::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s ( ", gcppcSLIDEReservedWords[sliderwtFace], 
	  m_pcID);

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      fprintf(fpOutputFile, "%s ", pCSLIDEPoint->GetID());      
    }

  fprintf(fpOutputFile, ")\n");

  if ( m_pcSurfaceID != gcppcSLIDEReservedWords[sliderwtSLF_INHERIT] )
    {
      fprintf(fpOutputFile, "  %s %s\n", 
	      gcppcSLIDEReservedWords[sliderwtSurface], 
	      m_pcSurfaceID);
    } 

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndFace]);
}

UINT_32 CSLIDEFace::GetPointCount()
{
  return m_listCSLIDEPoint.GetLength();
}

//////////////////////////////////////////////////////////////////////////////
// CSLIDEFaceIter Class
//

CSLIDEFaceIter::CSLIDEFaceIter()
{
}

CSLIDEFaceIter::~CSLIDEFaceIter()
{
}

VOID CSLIDEFaceIter::Init(CSLIDEFace *pCSLIDEFace)
{
  CSLIDEPrimitiveIter::Init(pCSLIDEFace);

  m_iter.Init(&(pCSLIDEFace->m_listCSLIDEPoint));
}

VOID CSLIDEFaceIter::Uninit()
{
  CSLIDEPrimitiveIter::Uninit();
}

CSLIDEPoint *CSLIDEFaceIter::PeekFirstCSLIDEPoint()
{
  return (CSLIDEPoint*)(m_iter.PeekFirst());
}

CSLIDEPoint *CSLIDEFaceIter::PeekNextCSLIDEPoint()
{
  return (CSLIDEPoint*)(m_iter.PeekNext());
}

CSLIDEPoint *CSLIDEFaceIter::PeekLastCSLIDEPoint()
{
  return (CSLIDEPoint*)(m_iter.PeekLast());
}

CSLIDEPoint *CSLIDEFaceIter::PeekPrevCSLIDEPoint()
{
  return (CSLIDEPoint*)(m_iter.PeekPrev());
}
